//
//  ExpressProxy.m
//  ZegoVoiceTranslation
//
//  Created by zego on 2022/10/28.
//

#import "ExpressProxy.h"
#import "RTCEngineManager.h"
#import "KeyCenter.h"
#import "ZegoSteamInfo.h"
@import ZegoRealtimeTranslation;

@interface ExpressProxy ()<ZegoEventHandler, ZegoAudioDataHandler, ZegoCustomAudioProcessHandler>
@property (nonatomic, weak) RTCEngineManager *rtcEngineManager;
@property (nonatomic, strong) ZegoExpressEngine *rtcEngine;
@property (nonatomic, assign) BOOL isInit;

@end

@implementation ExpressProxy

static NSString * const TAG = @"ExpressProxy";
- (instancetype)initWithEnginManager:(RTCEngineManager *)rtcEngineManager {
    self = [super init];
    if (self) {
        self.rtcEngineManager = rtcEngineManager;
    }
    return self;
}

- (void)initEngine {
    if (self.isInit) {
        return;
    }
    self.isInit = YES;
  
    if (!self.rtcEngine) {
        ZegoEngineProfile *engineProfile = [[ZegoEngineProfile alloc] init];
        engineProfile.appID = [KeyCenter appID];
        engineProfile.appSign = [KeyCenter appSign];
        engineProfile.scenario = ZegoScenarioDefault;
        
        self.rtcEngine = [ZegoExpressEngine createEngineWithProfile:engineProfile eventHandler:self];
        
        // Use the ZSE to process audio, so we disable the RTC's 3A ability
        [self.rtcEngine enableAEC:NO];
        [self.rtcEngine enableAGC:NO];
        [self.rtcEngine enableANS:NO];
        
        // Enable RTC Custom Audio Processing
        ZegoCustomAudioProcessConfig *config = [[ZegoCustomAudioProcessConfig alloc] init];
        config.sampleRate = ZegoAudioSampleRate16K;
        config.channel = ZegoAudioChannelMono;
        config.samples = 0.01 * 16000 * 1; // 10ms * 16K * Mono
        NSLog(@"⛓ Enable RTC custom audio capture processing");
        [self.rtcEngine enableCustomAudioCaptureProcessing:YES config:config];
        [self.rtcEngine setCustomAudioProcessHandler:self];

        // Use audio data observer to provide audio reference frames to AEC
        ZegoAudioFrameParam *param = [[ZegoAudioFrameParam alloc] init];
        param.sampleRate = ZegoAudioSampleRate16K;
        param.channel = ZegoAudioChannelMono;
        NSLog(@"⛓ Start RTC audio data observer");
        [self.rtcEngine startAudioDataObserver:ZegoAudioDataCallbackBitMaskPlayback|ZegoAudioDataCallbackBitMaskCaptured param:param];

        [self.rtcEngine setAudioDataHandler:self];
        
        ZegoVideoConfig *videoConfig = [[ZegoVideoConfig alloc] initWithPreset:ZegoVideoConfigPreset270P];
        videoConfig.captureResolution = CGSizeMake(1280, 720);
        videoConfig.encodeResolution = CGSizeMake(1280, 720);
        videoConfig.fps = 20;
        videoConfig.bitrate = 1500;
        [self.rtcEngine setVideoConfig:videoConfig];
    }
}

- (void)unInit:(void(^)(void))callback {
    [ZegoExpressEngine destroyEngine:^{
        callback();
    }];
}

- (void)login:(NSString *)roomId roomName:(NSString *)roomName userID:(NSString *)userID userName:(NSString *)userName callback:(void(^)(int errorCode, NSDictionary *extendedData))callback {
    // 创建用户对象，ZegoUser 的构造方法 initWithUserID “userID” 与 “userName” 不能为 “nil”，否则会导致登录房间失败。
    // userID 由您本地生成,需保证 “userID” 全局唯一。
    ZegoUser *user = [[ZegoUser alloc] initWithUserID:userID userName:userName];
    // 只有传入 “isUserStatusNotify” 参数取值为 “YES” 的 ZegoRoomConfig，才能收到 onRoomUserUpdate 回调。
    ZegoRoomConfig *config = [[ZegoRoomConfig alloc] init];
    config.isUserStatusNotify = YES;
    // 登录房间
    [self.rtcEngine loginRoom:roomId user:user config:config callback:^(int errorCode, NSDictionary * _Nonnull extendedData) {
        if (callback) {
            callback(errorCode, extendedData);
        }
    }];
}

- (void)logout:(NSString *)roomID {
    NSLog(@"%@ [logout]", TAG);
    [self.rtcEngine logoutRoom:roomID];
}

- (void)logout {
    NSLog(@"%@ [logout]", TAG);
    [self.rtcEngine logoutRoom];
}

- (void)startPublishStream:(NSString *)streamID {
    [self.rtcEngine startPublishingStream:streamID];
    ZegoAudioFrameParam *audioFrameParam = [[ZegoAudioFrameParam alloc] init];
    audioFrameParam.sampleRate = ZegoAudioSampleRate16K;
    audioFrameParam.channel = ZegoAudioChannelMono;
    [self.rtcEngine startAudioDataObserver:ZegoAudioDataCallbackBitMaskCaptured param:audioFrameParam];
}

- (void)stopPublishStream {
    [self.rtcEngine stopPublishingStream];
//    [self.engine stopPreview];
    [self.rtcEngine stopAudioDataObserver];
}

- (void)startPlayingStream:(NSString *)streamID view:(ZGView *)view {
    if (view != nil) {
        ZegoCanvas *canvas = [[ZegoCanvas alloc] initWithView:view];
        canvas.viewMode = ZegoViewModeAspectFill;
        [self.rtcEngine startPlayingStream:streamID canvas:canvas];
    } else {
        [self.rtcEngine startPlayingStream:streamID];
    }
}

- (void)stopPlayingStream:(NSString *)streamID {
    [self.rtcEngine stopPlayingStream:streamID];
}

- (void)startPreview:(ZGView *)view {
    ZegoCanvas *canvas = [[ZegoCanvas alloc] initWithView:view];
    canvas.viewMode = ZegoViewModeAspectFill;
    [self.rtcEngine startPreview:canvas];
}

- (void)sendSEIData:(NSData *)data {
    [self.rtcEngine sendSEI:data];
}


#pragma mark - ZegoEventHandler
- (void)onRoomStateUpdate:(ZegoRoomState)state errorCode:(int)errorCode extendedData:(NSDictionary *)extendedData roomID:(NSString *)roomID {
    NSLog(@"%@ [onRoomStateUpdate] %@, errorCode: %d roomState: %ld, state, extendedData: %@", TAG, roomID, errorCode, state, extendedData);
}

- (void)onRoomStateChanged:(ZegoRoomStateChangedReason)reason errorCode:(int)errorCode extendedData:(NSDictionary *)extendedData roomID:(NSString *)roomID {
    NSLog(@"%@ [onRoomStateChanged] reason: %lu, errorCode: %d, extendedData: %@", TAG, (unsigned long)reason, errorCode, extendedData);
    if (reason == ZegoRoomStateChangedReasonKickOut) {
        [self.rtcEngineManager onRoomStateChanged:RTCEngineRoomStateChangedReasonKickOut];
    } else if (reason == ZegoRoomStateChangedReasonReconnectFailed){
        [self.rtcEngineManager onRoomStateChanged:RTCEngineRoomStateChangedReasonDisconnect];
    }else if (reason == ZegoRoomStateChangedReasonLoginFailed) {
        [self.rtcEngineManager onRoomStateChanged:RTCEngineRoomStateChangedReasonLoginFailed];
    }else if (reason == ZegoRoomStateChangedReasonReconnected) {
        [self.rtcEngineManager onRoomStateChanged:RTCEngineRoomStateChangedReasonReconnected];
    }
}

- (void)onRoomStreamUpdate:(ZegoUpdateType)updateType streamList:(NSArray<ZegoStream *> *)streamList extendedData:(NSDictionary *)extendedData roomID:(NSString *)roomID {
    NSMutableArray<ZegoSteamInfo *> *streamInfoList = [NSMutableArray new];
    for (ZegoStream *stream in streamList) {
        NSLog(@"%@ [onRoomStreamUpdate] %@ %@", TAG, stream.streamID, updateType == ZegoUpdateTypeAdd ? @"add":@"remove");
        ZegoSteamInfo *streamInfo = [[ZegoSteamInfo alloc] initWithUserID:stream.user.userID userName:stream.user.userName streamID:stream.streamID];
        [streamInfoList addObject:streamInfo];
    }
    if (updateType == ZegoUpdateTypeAdd) {
        [self.rtcEngineManager onStreamAdd:streamInfoList];
    } else {
        [self.rtcEngineManager onStreamRemove:streamInfoList];
    }
}

- (void)onRoomStreamExtraInfoUpdate:(NSArray<ZegoStream *> *)streamList roomID:(NSString *)roomID {
    
}

- (void)onRoomExtraInfoUpdate:(NSArray<ZegoRoomExtraInfo *> *)roomExtraInfoList roomID:(NSString *)roomID {
    
}

- (void)onPlayerRecvSEI:(NSData *)data streamID:(NSString *)streamID {
    [self.rtcEngineManager onReceiveSEI:data streamID:streamID];
}

- (void)onPublisherStateUpdate:(ZegoPublisherState)state errorCode:(int)errorCode extendedData:(NSDictionary *)extendedData streamID:(NSString *)streamID {
    NSLog(@"%@ [onPublisherStateUpdate] streamID: %@ state: %lu errorCode: %d extendedData: %@", TAG, streamID, (unsigned long)state, errorCode, extendedData);
}

- (void)onPublisherQualityUpdate:(ZegoPublishStreamQuality *)quality streamID:(NSString *)streamID {
    NSLog(@"%@ onPublisherQualityUpdate:%@, %f, %f", TAG, streamID, quality.audioKBPS, quality.videoKBPS);
}

- (void)onPlayerQualityUpdate:(ZegoPlayStreamQuality *)quality streamID:(NSString *)streamID {
    NSLog(@"%@ onPlayerQualityUpdate:%@, %f, %f", TAG, streamID, quality.audioKBPS, quality.videoKBPS);
}

#pragma mark - ZegoAudioDataHandler
//- (void)onCapturedAudioData:(const unsigned char *)data dataLength:(unsigned int)dataLength param:(ZegoAudioFrameParam *)param {
//    // 收到本地采集的音频数据，开始推流后即可收到回调，该回调不在主线程中。开发者可以使用原始数据进行鉴黄等业务处理。
//    // The callback for obtaining the audio data captured by the local microphone.The callback is not in the main thread.Developers can use the original data to perform business processing such as pornographic identification.
//    [self.rtcEngineManager onCapturedAudioData:data dataLength:dataLength];
//}


#pragma mark - ZegoCustomAudioProcessHandler
- (void)onProcessCapturedAudioData:(unsigned char *)data dataLength:(unsigned int)dataLength param:(ZegoAudioFrameParam *)param timestamp:(double)timestamp {
    
    [self.rtcEngineManager onCapturedAudioData:data dataLength:dataLength];
}

@end
