//
//  RTCEngineManager.m
//  ZegoVoiceTranslation
//
//  Created by zego on 2022/10/28.
//

#import "RTCEngineManager.h"
#import <ZegoRealtimeTranslation/ZegoRealtimeTranslation.h>
#import "IRTCEventListener.h"

@interface RTCEngineManager ()
@property (nonatomic, copy) NSString *roomID;
@property (nonatomic, copy) NSString *userID;
@property (nonatomic, copy) NSString *userName;
@property (nonatomic, strong) NSMutableArray<ZegoSteamInfo *> *streamInfoList;
@property (nonatomic, strong) ExpressProxy *rtcEngine;
@property (nonatomic, assign) BOOL isInRoom;
@property (nonatomic, strong) NSMutableArray<id<IRTCEventListener>> *rtcEventListenerList;
@property (nonatomic, copy) NSString *publishStreamID;
@end

@implementation RTCEngineManager
static NSString * const TAG = @"RTCEngineManager";
NSString * const STREAM_PREFIX = @"zegotrans_";
static NSString * const SUBTITLE_RECOGNIZING_TEXT_PREFIX = @"zegoa2ti:";
static NSString * const SUBTITLE_RECOGNIZED_TEXT_PREFIX = @"zegoa2td:";

+ (instancetype)sharedManager {
    static RTCEngineManager *manager = nil;
    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        manager = [[RTCEngineManager alloc] init];
    });
    return manager;
}

- (instancetype)init
{
    self = [super init];
    if (self) {
        self.streamInfoList = [NSMutableArray new];
        self.rtcEventListenerList = [NSMutableArray new];
    }
    return self;
}

- (void)initRtc {
    self.rtcEngine = [[ExpressProxy alloc] initWithEnginManager:self];
    [self.rtcEngine initEngine];
}

- (void)unInitRtc:(void(^)(void))callback {
    NSLog(@"%@ [unInitRtc]", TAG);
    [self.rtcEngine unInit:^{
        callback();
    }];
    [self.streamInfoList removeAllObjects];
    [self.rtcEventListenerList removeAllObjects];
    self.rtcEngine = nil;
}

- (void)loginRoom:(NSString *)roomID roomName:(NSString *)roomName userID:(NSString *)userID userName:(NSString *)userName callback:(void(^)(int errorCode, NSDictionary *extendedData))callback {
    NSLog(@"%@ [loginRoom] roomID: %@, roomName: %@, userID: %@, userName: %@", TAG, roomID, roomName, userID, userName);
    self.roomID = roomID;
    self.userID = userID;
    self.userName = userName;
    if (self.rtcEngine == nil) {
        [self initRtc];
    }
    [self.rtcEngine login:roomID roomName:roomName userID:userID userName:userName callback:callback];
    self.isInRoom = YES;
}

- (void)logoutRoom:(NSString *)roomID {
    NSLog(@"%@ [logoutRoom]", TAG);
    if (self.isInRoom) {
        [self.rtcEngine logout:roomID];
        self.isInRoom = NO;
    }
}

- (void)logoutRoom {
    self.publishStreamID = @"";
    if (self.isInRoom) {
        [self.rtcEngine logout];
        self.isInRoom = NO;
    }
    // 退出时一定要把流信息列表清空
    [self.streamInfoList removeAllObjects];
}

- (void)startPublishStream:(NSString *)streamID {
    [self.rtcEngine startPublishStream:streamID];
}

- (void)stopPublishStream {
    [self.rtcEngine stopPublishStream];
}

- (void)startPlayingStream:(NSString *)streamID view:(UIView *)view {
    [self.rtcEngine startPlayingStream:streamID view:(ZGView *)view];
}

- (void)stopPlayingStream:(NSString *)streamID {
    [self.rtcEngine stopPlayingStream:streamID];
}

- (void)startPublishStream {
    self.publishStreamID = [NSString stringWithFormat:@"%@%@_%@", STREAM_PREFIX, self.roomID, self.userID];
    [self.rtcEngine startPublishStream:self.publishStreamID];
    NSMutableArray<NSString *> *streamIDList = [NSMutableArray new];
    for (ZegoSteamInfo *streamInfo in self.streamInfoList) {
        [streamIDList addObject:streamInfo.streamID];
    }
    if (![streamIDList containsObject:self.publishStreamID]) {
        ZegoSteamInfo *steamInfo = [[ZegoSteamInfo alloc] initWithUserID:self.userID userName:self.userName streamID:self.publishStreamID];
        [self.streamInfoList insertObject:steamInfo atIndex:0];
    }
}

- (void)startPreview:(UIView *)view {
    [self.rtcEngine startPreview:(ZGView *)view];
}

- (void)sendSEIData:(NSData *)data {
    [self.rtcEngine sendSEIData:data];
}

- (void)sendRecognizingText:(NSString *)text {
    NSString *content = [SUBTITLE_RECOGNIZING_TEXT_PREFIX stringByAppendingString:text];
    [self sendSEIData:[content dataUsingEncoding:NSUTF8StringEncoding]];
}

- (void)sendRecognizedText:(NSString *)text {
    NSString *content = [SUBTITLE_RECOGNIZED_TEXT_PREFIX stringByAppendingString:text];
    [self sendSEIData:[content dataUsingEncoding:NSUTF8StringEncoding]];
}

- (void)addRTCEventListener:(id<IRTCEventListener>)listener {
    if (![self.rtcEventListenerList containsObject:listener]) {
        [self.rtcEventListenerList addObject:listener];
    }
}

- (void)removeRTCEventListener:(id<IRTCEventListener>)listener {
    [self.rtcEventListenerList removeObject:listener];
}

#pragma mark - callback
- (void)onStreamAdd:(NSArray<ZegoSteamInfo *> *)streamInfoList {
    NSLog(@"%@ [onStreamAdd] add count: %ld, listenerList: %@",TAG, streamInfoList.count, self.rtcEventListenerList);
    [self.streamInfoList addObjectsFromArray:streamInfoList];
    // 流增加了，但是监听者还没有注册, 所以需要进入界面的时候遍历streamInfoList然后去更新UI
    for (id<IRTCEventListener> listener in self.rtcEventListenerList) {
        [listener onStreamAdd:streamInfoList];
    }
}

- (void)onStreamRemove:(NSArray<ZegoSteamInfo *> *)streamInfoList {
    [self.streamInfoList removeObjectsInArray:streamInfoList];
    for (id<IRTCEventListener> listener in self.rtcEventListenerList) {
        [listener onStreamRemove:streamInfoList];
    }
}

- (void)onRoomStateChanged:(RTCEngineRoomStateChangedReason)reason {
    if (reason == RTCEngineRoomStateChangedReasonReconnected) {  // 重连成功，重新推流
        if (self.publishStreamID.length > 0) {
            [self startPublishStream];
        }
    }else {
        // kickout | login failed | disconnect都要退出房间
        [self logoutRoom];
        for (id<IRTCEventListener> listener in self.rtcEventListenerList) {
            [listener onRoomStateChanged:reason];
        }
    }
    
}

- (void)onReceiveSEI:(NSData *)data streamID:(NSString *)streamID {
    NSString *str = [[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding];
    if ([str hasPrefix:SUBTITLE_RECOGNIZING_TEXT_PREFIX]) {
        NSString *subtitle = [str substringFromIndex:SUBTITLE_RECOGNIZING_TEXT_PREFIX.length];
        for (id<IRTCEventListener> listener in self.rtcEventListenerList) {
            [listener onRecognizingSEIText:subtitle];
        }
    } else if ([str hasPrefix:SUBTITLE_RECOGNIZED_TEXT_PREFIX]) {
        NSString *subtitle = [str substringFromIndex:SUBTITLE_RECOGNIZED_TEXT_PREFIX.length];
        for (id<IRTCEventListener> listener in self.rtcEventListenerList) {
            [listener onRecognizedSEIText:subtitle];
        }
    }
}

- (void)onCapturedAudioData:(const unsigned char *)data dataLength:(unsigned int)dataLength {
    if (self.isSpeechAudioFile) {
        return;
    }
    NSData *sendData = [NSData dataWithBytes:data length:dataLength];
    [[ZegoRealtimeTranslation getInstance] sendSpeechPCM:sendData];
}

@end
